__version__ = "0.0.11"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = '(detached from v0.0.11)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = 'b8d1705f26ed0388d3f9e0bb70b099943b9a0b24'
__git_head_subject__ = 'buildout.cfg: changed isolated python version to v2.7.6.6'
__git_head_message__ = ''
__git_dirty_diff__ = ''